=== modified file 'bzrlib/tests/test_transform.py'
--- bzrlib/tests/test_transform.py	2008-07-31 19:38:13 +0000
+++ bzrlib/tests/test_transform.py	2008-08-01 16:10:59 +0000
@@ -1277,6 +1277,27 @@
         self.assertEqual(wt.inventory.get_file_kind(wt.path2id("foo")),
                 "file")
 
+    def test_dir_to_hardlink(self):
+        wt = self.make_branch_and_tree('.')
+        self.build_tree(['foo/', 'foo/bar'])
+        wt.add(['foo', 'foo/bar'])
+        wt.commit("one")
+        tt = TreeTransform(wt)
+        self.addCleanup(tt.finalize)
+        foo_trans_id = tt.trans_id_tree_path("foo")
+        bar_trans_id = tt.trans_id_tree_path("foo/bar")
+        tt.delete_contents(foo_trans_id)
+        tt.delete_versioned(bar_trans_id)
+        self.build_tree(['baz'])
+        tt.create_hardlink("baz", foo_trans_id)
+        tt.apply()
+        self.failUnlessExists("foo")
+        self.failUnlessExists("baz")
+        wt.lock_read()
+        self.addCleanup(wt.unlock)
+        self.assertEqual(wt.inventory.get_file_kind(wt.path2id("foo")),
+                "file")
+
 
 class TransformGroup(object):
 

=== modified file 'bzrlib/transform.py'
--- bzrlib/transform.py	2008-07-31 19:38:13 +0000
+++ bzrlib/transform.py	2008-08-01 16:10:59 +0000
@@ -137,8 +137,6 @@
         self._case_sensitive_target = case_sensitive
         # A counter of how many files have been renamed
         self.rename_count = 0
-        # A set of the trans_ids that have changed kind
-        self._kind_change = set()
 
     def __get_root(self):
         return self._new_root
@@ -324,8 +322,6 @@
         New file takes the permissions of any existing file with that id,
         unless mode_id is specified.
         """
-        if not self._does_tree_kind_match(trans_id, 'file'):
-            self._kind_change.add(trans_id)
         name = self._limbo_name(trans_id)
         f = open(name, 'wb')
         try:
@@ -386,26 +382,11 @@
             os.unlink(name)
             raise
 
-    def _does_tree_kind_match(self, trans_id, kind):
-        """Checks whether the kind of trans_id has changed from the tree.
-
-        Returns True if trans_id exists in the tree and matches the
-        kind passed. Returns False otherwise.
-        """
-        try:
-            if self.tree_kind(trans_id) == kind:
-                return True
-        except NoSuchFile:
-            pass
-        return False
-
     def create_directory(self, trans_id):
         """Schedule creation of a new directory.
         
         See also new_directory.
         """
-        if not self._does_tree_kind_match(trans_id, 'directory'):
-            self._kind_change.add(trans_id)
         os.mkdir(self._limbo_name(trans_id))
         unique_add(self._new_contents, trans_id, 'directory')
 
@@ -415,8 +396,6 @@
         target is a bytestring.
         See also new_symlink.
         """
-        if not self._does_tree_kind_match(trans_id, 'symlink'):
-            self._kind_change.add(trans_id)
         if has_symlinks():
             os.symlink(target, self._limbo_name(trans_id))
             unique_add(self._new_contents, trans_id, 'symlink')
@@ -1238,10 +1217,10 @@
                 mover = _mover
             try:
                 child_pb.update('Apply phase', 0, 2)
-                self._apply_removals(new_inventory_delta, mover)
+                kind_changes = self._apply_removals(new_inventory_delta, mover)
                 child_pb.update('Apply phase', 1, 2)
                 modified_paths = self._apply_insertions(new_inventory_delta,
-                                                        mover)
+                                                        mover, kind_changes)
             except:
                 mover.rollback()
                 raise
@@ -1265,6 +1244,7 @@
         """
         tree_paths = list(self._tree_path_ids.iteritems())
         tree_paths.sort(reverse=True)
+        kind_changes = set()
         child_pb = bzrlib.ui.ui_factory.nested_progress_bar()
         try:
             for num, data in enumerate(tree_paths):
@@ -1272,6 +1252,12 @@
                 child_pb.update('removing file', num, len(tree_paths))
                 full_path = self._tree.abspath(path)
                 if trans_id in self._removed_contents:
+                    try:
+                        if (self.tree_kind(trans_id)
+                                != self.final_kind(trans_id)):
+                            kind_changes.add(trans_id)
+                    except NoSuchFile:
+                        pass
                     mover.pre_delete(full_path, os.path.join(self._deletiondir,
                                      trans_id))
                 elif trans_id in self._new_name or trans_id in \
@@ -1295,8 +1281,9 @@
                     inventory_delta.append((path, None, file_id, None))
         finally:
             child_pb.finished()
+        return kind_changes
 
-    def _apply_insertions(self, inventory_delta, mover):
+    def _apply_insertions(self, inventory_delta, mover, kind_changes):
         """Perform tree operations that insert directory/inventory names.
 
         That is, create any files that need to be created, and restore from
@@ -1305,6 +1292,9 @@
 
         If inventory_delta is None, no inventory delta is calculated, and
         no list of modified paths is returned.
+
+        kind_changes is a set of trans ids where the entry has changed
+        kind, and so an inventory delta entry should be created for them.
         """
         new_paths = self.new_paths(filesystem_only=(inventory_delta is None))
         modified_paths = []
@@ -1342,7 +1332,7 @@
                         trans_id in self._new_name or
                         trans_id in self._new_parent
                         or trans_id in self._new_executability
-                        or trans_id in self._kind_change):
+                        or trans_id in kind_changes):
                         try:
                             kind = self.final_kind(trans_id)
                         except NoSuchFile:

